﻿using Styx;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

#region [Method] - Class Redundancy
using L = MonkzShaman.Core.Utilities.Log;
using SB = MonkzShaman.Core.Helpers.Spell_Book;
using TM = MonkzShaman.Core.Managers.Talent_Manager;
using MAIN = MonkzShaman;
using System.Collections;
using MonkzShaman.Core;
using System.Diagnostics;
#endregion

namespace MonkzShaman.Core
{
    static class Unit
    {

        #region [Method] - Active Enemies
        private static LocalPlayer Me { get { return Styx.StyxWoW.Me; } }
        private static WoWUnit currentTarget { get { return StyxWoW.Me.CurrentTarget; } }

        public static IEnumerable<WoWUnit> UnfriendlyUnits
        {
            get
            {
                //allowInheritance  [True] would also return WoWPlayers.
                return ObjectManager.GetObjectsOfType<WoWUnit>(allowInheritance: true, includeMeIfFound: true).Where(u => u != null && u.IsValid && !u.IsDead && u.Attackable && !u.IsCritter);
            }
        }


        public static int enemiesInRangeOfTarget(int range)
        {
            return UnfriendlyUnits.Count(u => u.Combat && u.Location.Distance(currentTarget.Location) - u.CombatReach <= range && u.CurrentHealth > 1);
        }

        public static int enemiesInRangeOfMe(int range)
        {
            return UnfriendlyUnits.Count(u => u.Combat && u.Location.Distance(Me.Location) - u.CombatReach <= range && u.CurrentHealth > 1);
        }

        #endregion

        #region [Method] - Aura Detection
        public static bool auraExists(this WoWUnit Unit, int auraID, bool isMyAura = false)
        {
            try
            {
                if (Unit == null || !Unit.IsValid)
                    return false;
                WoWAura Aura = isMyAura ? Unit.GetAllAuras().FirstOrDefault(A => A.SpellId == auraID && A.CreatorGuid == Me.Guid) : Unit.GetAllAuras().FirstOrDefault(A => A.SpellId == auraID);
                return Aura != null;
            }
            catch (Exception xException)
            {
                L.diagnosticLog("Exception in auraExists(): ", xException);
                return false;
            }
        }

        public static uint auraStacks(this WoWUnit Unit, int auraID, bool isMyAura = false)
        {
            try
            {
                if (Unit == null || !Unit.IsValid)
                    return 0;
                WoWAura Aura = isMyAura ? Unit.GetAllAuras().FirstOrDefault(A => A.SpellId == auraID && A.CreatorGuid == Me.Guid) : Unit.GetAllAuras().FirstOrDefault(A => A.SpellId == auraID);
                return Aura != null ? Aura.StackCount : 0;
            }
            catch (Exception xException)
            {
                L.diagnosticLog("Exception in auraStacks(): ", xException);
                return 0;
            }
        }

        public static double auraTimeLeft(this WoWUnit Unit, int auraID, bool isMyAura = false)
        {
            try
            {
                if (Unit == null || !Unit.IsValid)
                    return 0;
                WoWAura Aura = isMyAura ? Unit.GetAllAuras().FirstOrDefault(A => A.SpellId == auraID && A.CreatorGuid == Me.Guid) : Unit.GetAllAuras().FirstOrDefault(A => A.SpellId == auraID);
                return Aura != null ? Aura.TimeLeft.TotalSeconds : 0;
            }
            catch (Exception xException)
            {
                L.diagnosticLog("Exception in auraTimeLeft(): ", xException);
                return 9999;
            }
        }

        #endregion

        public static Boolean isBloodlusted()
        {
            if (auraExists(Me, 2825) || auraExists(Me, 80354) || auraExists(Me, 32182))
            {
                return true;
            }
            return false;
        }

        //#region[Method] Timespan
        //public static void startTimer()
        //{
        //    timerStart = DateTime.Now;
        //    pause = true;
        //}
        //#endregion

        #region [Method] - Cache
        public static DateTime timerStart;
        public static DateTime combatStart;

        //public static uint energy_time_to_max;
        public static uint ls_stacks;
        public static double flameshock_time_left;
        public static bool flameshock_active;
        public static bool ascendance;
        public static int combatTime;
        public static bool openerTime;

        //public static void pauseCasting()
        //{
        //    castPauseStart = DateTime.Now;
        //    castPause = true;
        //}

        public static void Cache()
        {

            ls_stacks = auraStacks(Me, SB.aurals_charges);
            flameshock_time_left = auraTimeLeft(currentTarget, SB.flame_shock, true);
            ascendance = auraExists(Me, SB.ascendance);
            if (flameshock_time_left > 0)
            {
                flameshock_active = true;
            }
            else
            {
                flameshock_active = false;
            }
            
            //if (castPause)
            //{
            //    TimeSpan span = DateTime.Now - castPauseStart;
            //    int pauseRunning = (int)span.TotalMilliseconds;
            //    if (pauseRunning > 200)
            //    {
            //        castPause = false;
            //        L.diagnosticLog("Casting Resuming");

            //    }
            //}
            
            /**if (pause)
            //{
            //    TimeSpan span = DateTime.Now - timerStart;
            //    int pauseRunning = (int)span.TotalMilliseconds;
            //    if (pauseRunning > 3000)
            //    {
            //        pause = false;
            //        L.combatLog("CR Resuming");

            //    }
            //}**/

            

        }
        #endregion

        #region [Method] - Customized Singular Totem Function
        public static bool IsRealTotem(WoWTotem ti)
        {
            return ti != WoWTotem.None
                && ti != WoWTotem.DummyAir
                && ti != WoWTotem.DummyEarth
                && ti != WoWTotem.DummyFire
                && ti != WoWTotem.DummyWater;
        }

        /// <summary>
        /// check if a specific totem (ie Mana Tide Totem) exists
        /// </summary>
        /// <param name="wtcheck"></param>
        /// <returns></returns>
        public static bool Exist(WoWTotem wtcheck)
        {
            WoWTotemInfo tiexist = GetTotem(wtcheck);
            return tiexist != null;
        }

        /// <summary>
        /// check if a WoWTotemInfo object references a real totem (other than None or Dummies)
        /// </summary>
        /// <param name="ti"></param>
        /// <returns></returns>
        public static bool Exist(WoWTotemInfo ti)
        {
            return IsRealTotem(ti.WoWTotem);
        }

        /// <summary>
        /// check if a type of totem (ie Air Totem) exists
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        public static bool Exist(WoWTotemType type)
        {
            WoWTotem wt = GetTotem(type).WoWTotem;
            return IsRealTotem(wt);
        }

        /// <summary>
        /// check if any of several specific totems exist
        /// </summary>
        /// <param name="wt"></param>
        /// <returns></returns>
        public static bool Exist(params WoWTotem[] wt)
        {
            return wt.Any(t => Exist(t));
        }

        /// <summary>
        /// check if a specific totem exists within its max range of a given location
        /// </summary>
        /// <param name="pt"></param>
        /// <param name="tt"></param>
        /// <returns></returns>
        public static bool ExistInRange(WoWPoint pt, WoWTotem tt)
        {
            if (!Exist(tt))
                return false;

            WoWTotemInfo ti = GetTotem(tt);
            return ti.Unit != null && ti.Unit.Location.Distance(pt) < GetTotemRange(tt);
        }

        /// <summary>
        /// check if any of several totems exist in range
        /// </summary>
        /// <param name="pt"></param>
        /// <param name="awt"></param>
        /// <returns></returns>
        public static bool ExistInRange(WoWPoint pt, params WoWTotem[] awt)
        {
            return awt.Any(t => ExistInRange(pt, t));
        }

        /// <summary>
        /// check if type of totem (ie Air Totem) exists in range
        /// </summary>
        /// <param name="pt"></param>
        /// <param name="type"></param>
        /// <returns></returns>
        public static bool ExistInRange(WoWPoint pt, WoWTotemType type)
        {
            WoWTotemInfo ti = GetTotem(type);
            return Exist(ti) && ti.Unit != null && ti.Unit.Location.Distance(pt) < GetTotemRange(ti.WoWTotem);
        }
        #endregion 

        /// <summary>
        /// gets reference to array element in Me.Totems[] corresponding to WoWTotemType of wt.  Return is always non-null and does not indicate totem existance
        /// </summary>
        /// <param name="wt">WoWTotem of slot to reference</param>
        /// <returns>WoWTotemInfo reference</returns>
        public static WoWTotemInfo GetTotem(WoWTotem wt)
        {
            WoWTotemInfo ti = GetTotem(wt.ToType());
            if (ti.WoWTotem != wt)
                return null;
            return ti;
        }

        /// <summary>
        /// gets reference to array element in Me.Totems[] corresponding to type.  Return is always non-null and does not indicate totem existance
        /// </summary>
        /// <param name="type">WoWTotemType of slot to reference</param>
        /// <returns>WoWTotemInfo reference</returns>
        public static WoWTotemInfo GetTotem(WoWTotemType type)
        {
            return Me.Totems[(int)type - 1];
        }

        public static double GetTotemTimeLeft(WoWTotemType type)
        {
            TimeSpan running = Me.Totems[(int)type - 1].Expires.Subtract(DateTime.Now);
            //L.diagnosticLog("" + Me.Totems[(int)type - 1].Name + " time left: " + running.TotalSeconds);
            return running.TotalSeconds;
        }

        /// <summary>
        /// check if all totems are within range of Shaman's location
        /// </summary>
        public static int TotemsInRange 
        { 
            get 
            {
                return TotemsInRangeOf(StyxWoW.Me);
            }
        }

        /// <summary>
        /// check if all totems are within range of a WoWUnit
        /// </summary>
        /// <param name="unit"></param>
        /// <returns></returns>
        public static int TotemsInRangeOf(WoWUnit unit)
        {
            return StyxWoW.Me.Totems.Where(t => t.Unit != null).Count(t => unit.Location.Distance(t.Unit.Location) < GetTotemRange(t.WoWTotem));
        }

        /// <summary>
        ///   Finds the max range of a specific totem, where you'll still receive the buff.
        /// </summary>
        /// <remarks>
        ///   Created 3/26/2011.
        /// </remarks>
        /// <param name = "totem">The totem.</param>
        /// <returns>The calculated totem range.</returns>
        public static float GetTotemRange(WoWTotem totem)
        {
            switch (totem)
            {
                case WoWTotem.Tremor:
                    return 30f;

                case WoWTotem.Searing:
                    if ( MAIN.Main.spec == WoWSpec.ShamanElemental )
                        return 40;
                    return 20f;

                case WoWTotem.Earthbind:
                    return 10f;

                case WoWTotem.Grounding:
                case WoWTotem.Magma:
                    return 8f;

                case WoWTotem.EarthElemental:
                case WoWTotem.FireElemental:
                    // Not really sure about these 3.
                    return 20f;

                case WoWTotem.ManaTide:
                    // Again... not sure :S
                    return 40f;


                case WoWTotem.Earthgrab:
                    return 10f;

                case WoWTotem.StoneBulwark:
                    // No idea, unlike former glyphed stoneclaw it has a 5 sec pluse shield component so range is more important
                    return 40f;

                case WoWTotem.HealingStream:
                    return 40f;

                case WoWTotem.HealingTide:
                    return 40f;

                case WoWTotem.Capacitor:
                    return 8f;

                case WoWTotem.Windwalk:
                    return 40f;

                case WoWTotem.SpiritLink:
                    return 10f;
            }

            return 0f;
        }


        public static int ToSpellId(this WoWTotem totem)
        {
            return (int)(((long)totem) & ((1L << 32) - 1));
        }

        public static WoWTotemType ToType(this WoWTotem totem)
        {
            return (WoWTotemType)((long)totem >> 32);
        }


        static WoWTotem[] totemsWeDontRecall = new WoWTotem[] 
        {
            WoWTotem.FireElemental , 
            WoWTotem.EarthElemental  , 
            WoWTotem.HealingTide , 
            WoWTotem.ManaTide 
        };

        static WoWTotem[] totemsThatAggro = new WoWTotem[] 
        {
            WoWTotem.FireElemental,
            WoWTotem.Searing,
            WoWTotem.Magma,
            WoWTotem.EarthElemental,
            WoWTotem.Earthbind,
            WoWTotem.Earthgrab,
            WoWTotem.Capacitor,
             
        };

        public static bool IsSearingTotemNeeded()
        {
                WoWTotemInfo ti = GetTotem(WoWTotemType.Fire);
                if (ti.WoWTotem == WoWTotem.FireElemental)
                {
                    return false;
                }
                if (!Exist(ti))
                {
                    L.diagnosticLog("IsSearingTotemNeeded: no Fire Totem so setting Searing Totem");
                    return true;
                }

                if (ti.WoWTotem == WoWTotem.Searing)
                {
                    float currDist = ti.Unit.Location.Distance(Me.CurrentTarget.Location);
                    currDist = currDist - Me.CurrentTarget.CombatReach;
                    if (currDist > GetTotemRange(WoWTotem.Searing))
                    {
                        L.diagnosticLog("IsSearingTotemNeeded: Searing Totem is {0:F1} yds from CurrentTarget, refreshing on "+ Me.CurrentTarget.SafeName, currDist);
                        return true;
                    }
                }

               
            return false;
        }

        #region [Method] - Unit Evaluation
        public static bool isUnitValid(this WoWUnit Unit, double spellRange)
        {
            return Unit != null && Unit.IsValid && Unit.IsAlive && Unit.Attackable && Unit.Distance <= Unit.CombatReach + spellRange;
        }

        #endregion

        #region [Method] - Unit Evaluation
        public static bool isUnitValid(this WoWUnit Unit, double spellRange, bool MeleeRange = false)
        {
            if (MeleeRange == true)
                return Unit != null && Unit.IsValid && Unit.IsAlive && Unit.Attackable && Me.IsWithinMeleeRangeOf(Unit);
            else
            {
                return Unit != null && Unit.IsValid && Unit.IsAlive && Unit.Attackable && Unit.Distance <= Unit.CombatReach + spellRange;
            }
        }
        #endregion

    }
}